package jp.kirikiri.tvp2.visual;

import jp.kirikiri.tjs2.TJSException;

public interface DivisibleTransHandler extends BaseTransHandler {

	/**
	 * called before one processing time unit.
	 * expected return values are:
	 * S_TRUE: continue processing
	 * S_FALSE: break processing
	 * @param tick
	 * @return
	 */
	public int startProcess( /*in*/long tick );

	/**
	 * called after one processing time unit.
	 * expected return values are:
	 * S_TRUE: continue processing
	 * S_FALSE: break processing
	 */
	public int endProcess();

	/**
	 * called during StartProcess and EndProcess per an update rectangle.
	 * the handler processes given rectangle and put result image to
	 * "Dest"( in tTVPDivisibleData ).
	 * given "Dest" is a internal image buffer, but callee can change
	 * the "Dest" pointer to Src1 or Src2. Also DestLeft and DestTop can
	 * be changed to point destination image part.
	 */
	public int process( /*in,out*/DivisibleData data );

	/**
	 * will be called after StartProcess/EndProcess returns S_FALSE.
	 * this function does not called in some occasions.
	 * fill "dest" to make a final image.
	 * dest can be set to either src1 or src2.
	 * @throws TJSException
	 */
	public int makeFinalImage( /*in,out*/ScanLineProvider dest, /*in*/ScanLineProvider src1, /*in*/ScanLineProvider src2 ) throws TJSException;
}
